@@ -1,9 +1,9 @@
 /*****
 *
-* Copyright (C) 2001, 2002 Yoann Vandoorselaere <yoann@prelude-ids.org>
-* All Rights Reserved
+* Copyright (C) 2001, 2002, 2003, 2004, 2005 PreludeIDS Technologies. All Rights Reserved.
+* Author: Yoann Vandoorselaere <yoann.v@prelude-ids.com>
 *
-* This file is part of the Prelude program.
+* This file is part of the Prelude-Manager program.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
@@ -24,10 +24,16 @@
 #ifndef _MANAGER_SENSOR_SERVER_H
 #define _MANAGER_SENSOR_SERVER_H
 
-void sensor_server_close(server_generic_t *server);
+void sensor_server_stop(server_generic_t *server);
 
-server_generic_t *sensor_server_new(const char *addr, uint16_t port);
+server_generic_t *sensor_server_new(void);
 
 int sensor_server_broadcast_admin_command(uint64_t *analyzerid, prelude_msg_t *msg);
 
+int sensor_server_add_client(server_generic_t *server, prelude_connection_t *cnx);
+
+int sensor_server_write_client(server_generic_client_t *dst, prelude_msg_t *msg);
+
+int sensor_server_queue_write_client(server_generic_client_t *client, void *msg);
+
 #endif /* _MANAGER_SENSOR_SERVER_H */
