@@ -1,9 +1,9 @@
 /*****
 *
-* Copyright (C) 1999,2000, 2002, 2003 Yoann Vandoorselaere <yoann@prelude-ids.org>
-* All Rights Reserved
+* Copyright (C) 1999, 2000, 2001, 2002, 2003, 2004, 2005 PreludeIDS Technologies. All Rights Reserved.
+* Author: Yoann Vandoorselaere <yoann.v@prelude-ids.com>
 *
-* This file is part of the Prelude program.
+* This file is part of the Prelude-Manager program.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by 
@@ -21,7 +21,11 @@
 *
 *****/
 
+#include "config.h"
+#include "libmissing.h"
+
 #include <stdio.h>
+#include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
@@ -36,230 +40,85 @@
 #include <sys/poll.h>
 #include <assert.h>
 #include <sys/stat.h>
+#include <signal.h>
 
 #include <libprelude/common.h>
 #include <libprelude/prelude-log.h>
-#include <libprelude/config-engine.h>
-#include <libprelude/plugin-common.h>
 #include <libprelude/prelude-io.h>
-#include <libprelude/prelude-message.h>
+#include <libprelude/prelude-msg.h>
 #include <libprelude/prelude-message-id.h>
-#include <libprelude/prelude-auth.h>
-#include <libprelude/prelude-path.h>
-#include <libprelude/extract.h>
-#include <libprelude/prelude-inet.h>
+#include <libprelude/prelude-client.h>
+#include <libprelude/prelude-error.h>
 
-#include "config.h"
-#include "ssl.h"
-#include "auth.h"
-#include "pconfig.h"
+#include <gnutls/gnutls.h>
+
+#include "manager-auth.h"
 #include "server-logic.h"
 #include "server-generic.h"
 
 
-
 struct server_generic {
 
         int sock;
 
-        size_t sa_len;
+        size_t slen;
         struct sockaddr *sa;
         
-        
         size_t clientlen;
         struct server_logic *logic;
         server_generic_read_func_t *read;
+        server_generic_write_func_t *write;
         server_generic_close_func_t *close;
         server_generic_accept_func_t *accept;
 };
 
 
-
 struct server_generic_client {
         SERVER_GENERIC_OBJECT;
 };
 
 
 
-
-/*
- * Generate a configuration message containing
- * the kind of connection the Manager support.
- */
-static prelude_msg_t *generate_config_message(void)
-{
-        prelude_msg_t *msg;
-
-        msg = prelude_msg_new(2, 0, PRELUDE_MSG_AUTH, 0);
-        if ( ! msg )
-                return NULL;
-        
-#ifdef HAVE_SSL
-        prelude_msg_set(msg, PRELUDE_MSG_AUTH_HAVE_SSL, 0, NULL);
-#endif
-        prelude_msg_set(msg, PRELUDE_MSG_AUTH_HAVE_PLAINTEXT, 0, NULL);
-        
-        return msg;
-}
+extern prelude_client_t *manager_client;
+static volatile sig_atomic_t continue_processing = 1;
 
 
 
 
-static int send_plaintext_authentication_result(prelude_io_t *fd, uint8_t tag)
+
+static int send_auth_result(server_generic_client_t *client, int result)
 {
         int ret;
-        prelude_msg_t *msg;
+        uint64_t nident;
+        prelude_client_profile_t *cp;
         
-        msg = prelude_msg_new(1, 0, PRELUDE_MSG_AUTH, 0);
-        if ( ! msg )
-                return -1;
-        
-        prelude_msg_set(msg, tag, 0, NULL);
-        ret = prelude_msg_write(msg, fd);
-        prelude_msg_destroy(msg);
-
-        return ret;
-}
-
-
-
-
-/*
- * Start the SSL authentication process.
- */
-static prelude_msg_status_t handle_ssl_authentication(server_generic_client_t *client) 
-{
-#ifdef HAVE_SSL
-        int ret;
-        
-        ret = ssl_auth_client(client->fd);
-        if ( ret < 0 ) {
-                log(LOG_INFO, "[%s] - SSL authentication failed.\n", client->addr);
-                return -1;
-        }
-        
-        if ( ret == 0 )
-                /*
-                 * unfinished because of non blocking.
-                 */
-                return 0;
-
-        client->is_authenticated = 1;
-        log(LOG_INFO, "[%s] - SSL authentication succeed.\n", client->addr);
-        
-        return 1;
-#else
-        log(LOG_INFO, "[%s] - requested unavailable option : SSL.\n");
-        return -1;
-#endif
-}
-
-
-
-
-/*
- * Read authentication information contained in the passed message.
- * Then try to authenticate the peer.
- */
-static int handle_plaintext_authentication(prelude_msg_t *msg, server_generic_client_t *client)
-{
-        int ret;
-        void *buf;
-        uint8_t tag;
-        uint32_t len;
-        const char *user = NULL, *pass = NULL;
-
-        while ( (ret = prelude_msg_get(msg, &tag, &len, &buf)) > 0 ) {
-                
-                switch (tag) {
-                        
-                case PRELUDE_MSG_AUTH_USERNAME:
-                        ret = extract_string_safe(&user, buf, len);
-                        if ( ret < 0 )
-                                return -1;
-                        break;
-                        
-                case PRELUDE_MSG_AUTH_PASSWORD:
-                        ret = extract_string_safe(&pass, buf, len);
-                        if ( ret < 0 )
-                                return -1;
-                        break;
-                        
-                default:
-                        log(LOG_INFO, "[%s] - invalid authentication message.\n", client->addr);
+        if ( ! client->msg ) {
+                ret = prelude_msg_new(&client->msg, 1, sizeof(uint64_t), PRELUDE_MSG_AUTH, 0);
+                if ( ret < 0 )
                         return -1;
-                }
-        }
-        
-        if ( ! user || ! pass || ret < 0 ) {
-                log(LOG_INFO, "[%s] - invalid authentication message.\n", client->addr);
-                return -1;
-        }
-        
-        ret = prelude_auth_check(MANAGER_AUTH_FILE, user, pass);
-        if ( ret < 0 ) {
-                log(LOG_INFO, "[%s] - plaintext authentication failed.\n", client->addr);
-                send_plaintext_authentication_result(client->fd, PRELUDE_MSG_AUTH_FAILED);
-                return -1;
-        }
-
-        log(LOG_INFO, "[%s] - plaintext authentication succeed.\n", client->addr);
-
-        ret = send_plaintext_authentication_result(client->fd, PRELUDE_MSG_AUTH_SUCCEED);
-        if ( ret < 0 ) {
-                log(LOG_ERR, "error sending authentication result.\n");
-                return -1;
-        }
-        
-        client->is_authenticated = 1;
-        
-        return 0;
-}
-
-
-
-
-/*
- * Either plaintext, either SSL.
- * call the necessary authentication function.
- */
-static int handle_authentication(prelude_msg_t *msg, server_generic_client_t *client) 
-{
-        int ret;
-        void *buf;
-        uint8_t tag;
-        uint32_t dlen;
-        
-        tag = prelude_msg_get_tag(msg);
-
-        if ( tag != PRELUDE_MSG_AUTH ) {
-                log(LOG_INFO, "[%s] - expected authentication tag got (%d).\n", client->addr, tag);
-                return -1;
-        }
-
-        ret = prelude_msg_get(msg, &tag, &dlen, &buf);
-        if ( ret <= 0 )
-                return -1;
-        
-        switch (tag) {
-
-        case PRELUDE_MSG_AUTH_SSL:
-                client->is_ssl = 1;
-                ret = handle_ssl_authentication(client);
-                break;
                 
-        case PRELUDE_MSG_AUTH_PLAINTEXT:
-                client->is_ssl = 0;
-                ret = handle_plaintext_authentication(msg, client);                
-                break;
-
-        default:
-                log(LOG_INFO, "[%s] - invalid authentication tag (%d).\n", client->addr, tag);
-                return -1;
+                cp = prelude_client_get_profile(manager_client);
+                nident = prelude_hton64(prelude_client_profile_get_analyzerid(cp));
+                prelude_msg_set(client->msg, result, sizeof(nident), &nident);
+        }
+        
+        ret = prelude_msg_write(client->msg, client->fd);
+        
+        if ( ret < 0 ) {
+		if ( prelude_error_get_code(ret) == PRELUDE_ERROR_EAGAIN ) {
+			server_logic_notify_write_enable((server_logic_client_t *) client);
+			return 0;
+		}
+                
+		prelude_msg_destroy(client->msg);
+		return -1;
         }
 
+        prelude_msg_destroy(client->msg);
 
-        return ret;
+        client->msg = NULL;
+                
+        return (client->state & SERVER_GENERIC_CLIENT_STATE_AUTHENTICATED) ? 1 : -1;
 }
 
 
@@ -275,46 +134,58 @@ static int handle_authentication(prelude_msg_t *msg, server_generic_client_t *cl
 static int authenticate_client(server_generic_t *server, server_generic_client_t *client) 
 {
         int ret;
-        prelude_msg_status_t status;
         
-        if ( client->is_ssl == 1 ) {
-                /*
-                 * FIXME:
-                 * 
-                 * handle_authentication() was previously called, and it was an
-                 * SSL kind of connection. Don't try to read a prelude-message:
-                 * directly call the SSL subsystem, so that we can finish
-                 * authenticating the connection.
-                 *
-                 * This is a hack, and using prelude-message for SSL authentication
-                 * would be a good thing (if possible).
-                 */
-                ret = handle_ssl_authentication(client);
-                if ( ret <= 0 )
-                        return ret;
+        if ( ! client->msg && ! (client->state & SERVER_GENERIC_CLIENT_STATE_AUTHENTICATED) ) {
+
+                ret = manager_auth_client(client, client->fd);                
+                if ( ret == 0 )
+                        return ret; /* EAGAIN happened */
                 
-                return server->accept(client);
+                if ( ret < 0 )
+                        return send_auth_result(client, PRELUDE_MSG_AUTH_FAILED);
+                
+                client->state |= SERVER_GENERIC_CLIENT_STATE_AUTHENTICATED;
+                
+                ret = send_auth_result(client, PRELUDE_MSG_AUTH_SUCCEED);
+                if ( ret != 1 )
+                        return ret;
         }
 
-        else {
-                status = prelude_msg_read(&client->msg, client->fd);        
-
-                if ( status == prelude_msg_finished ) {        
-                        ret = handle_authentication(client->msg, client);
-                        
-                        prelude_msg_destroy(client->msg);
-                        client->msg = NULL;
-                        
-                        if ( ret < 0 || (ret == 0 && client->is_ssl) )
-                                return ret;
-
-                        return server->accept(client);
-                }
-                
-                else if ( status == prelude_msg_unfinished )
-                        return 0;
-                
+        else if ( client->msg ) {
+                ret = send_auth_result(client, -1);
+                if ( ret != 1 )
+                        return ret;
+        }
+        
+        if ( ! client->state & SERVER_GENERIC_CLIENT_STATE_AUTHENTICATED )
                 return -1;
+        
+        if ( server->sa->sa_family == AF_UNIX && ! (client->state & SERVER_GENERIC_CLIENT_STATE_ACCEPTED) ) {
+                ret = manager_auth_disable_encryption(client, client->fd);
+                if ( ret <= 0 )
+                        return ret;
+
+                server_generic_log_client(client, PRELUDE_LOG_INFO, "disabled encryption on local UNIX connection.\n");
+        }
+
+        client->state |= SERVER_GENERIC_CLIENT_STATE_ACCEPTED;
+        
+        return server->accept(client);
+}
+
+
+
+
+static int write_connection_cb(void *sdata, server_logic_client_t *ptr)
+{
+        server_generic_t *server = sdata;
+        server_generic_client_t *client = (server_generic_client_t *) ptr;
+        
+        if ( client->state & SERVER_GENERIC_CLIENT_STATE_ACCEPTED )
+                return server->write(client);
+        else {
+                server_logic_notify_write_disable(ptr);
+                return authenticate_client(sdata, (server_generic_client_t *) ptr);
         }
 }
 
@@ -334,15 +205,18 @@ static int read_connection_cb(void *sdata, server_logic_client_t *ptr)
         int ret = 0;
         server_generic_t *server = sdata;
         server_generic_client_t *client = (server_generic_client_t *) ptr;
-
-        do {    
-                if ( client->is_authenticated )
-                        ret = server->read(client);
-                else
-                        ret = authenticate_client(server, client);
-
-        } while ( ret == 0 && prelude_io_pending(client->fd) > 0 );
-
+        
+        if ( client->state & SERVER_GENERIC_CLIENT_STATE_CLOSING ) {
+                /* stop further processing */
+                ret = (server_logic_remove_client(ptr) == 0) ? -2 : 0;
+        }
+        
+        else if ( client->state & SERVER_GENERIC_CLIENT_STATE_ACCEPTED )
+                ret = server->read(client);
+        
+        else
+                ret = authenticate_client(server, client);
+        
         return ret;
 }
 
@@ -355,20 +229,44 @@ static int read_connection_cb(void *sdata, server_logic_client_t *ptr)
  */
 static int close_connection_cb(void *sdata, server_logic_client_t *ptr) 
 {
+        int ret;
         server_generic_t *server = sdata;
         server_generic_client_t *client = (server_generic_client_t *) ptr;
-        
-        log(LOG_INFO, "[%s] - closing connection.\n", client->addr);
 
-        free(client->addr);
+        client->state |= SERVER_GENERIC_CLIENT_STATE_CLOSING;
         
-        prelude_io_close(client->fd);
-        prelude_io_destroy(client->fd);
+        if ( client->state & SERVER_GENERIC_CLIENT_STATE_ACCEPTED && ! (client->state & SERVER_GENERIC_CLIENT_STATE_CLOSED) ) {
+                
+                ret = server->close(client);
+                if ( ret < 0 )
+                        return ret;
+
+                client->state |= SERVER_GENERIC_CLIENT_STATE_CLOSED;
+        }
         
-        if ( client->is_authenticated )
-                server->close(client);
+        /*
+         * layer above server-generic are permited to set fd to NULL so
+         * that they can take control over the connection FD.
+         */
+        if ( client->fd ) {
+                
+                ret = prelude_io_close(client->fd);
+                if ( ret < 0 && prelude_error_get_code(ret) == PRELUDE_ERROR_EAGAIN ) {
+
+                        if ( server->sa->sa_family != AF_UNIX && gnutls_record_get_direction(prelude_io_get_fdptr(client->fd)) == 1 )
+                                server_logic_notify_write_enable(ptr);
+                        
+                        return -1;
+                }
+                
+                prelude_io_destroy(client->fd);
+        }
         
-        free(ptr);
+        server_generic_log_client(client, PRELUDE_LOG_INFO, "closing connection.\n");
+
+        free(client->permission_string);
+        free(client->addr);        
+        free(client);
         
         return 0;
 }
@@ -398,11 +296,9 @@ static int tcpd_auth(server_generic_client_t *cdata, int clnt_sock)
 
         ret = hosts_access(&request);
         if ( ! ret ) {
-                log(LOG_INFO, "[%s] - tcp wrapper refused connection.\n", cdata->addr);
+                server_generic_log_client(cdata, PRELUDE_LOG_WARN, "tcp wrapper refused connection.\n", cdata->addr);
                 return -1;
         }
-
-        log(LOG_INFO, "[%s] - tcp wrapper accepted connection.\n", cdata->addr);
         
         return 0;
 }
@@ -428,29 +324,27 @@ static int setup_client_socket(server_generic_t *server,
                 ret = tcpd_auth(cdata, client);
                 if ( ret < 0 )
                         return -1;
-        } else
-                log(LOG_INFO, "[%s] - accepted connection.\n", cdata->addr);
-#else
-        log(LOG_INFO, "[%s] - accepted connection.\n", cdata->addr);
+        }
 #endif
         /*
          * set client socket non blocking.
          */
         ret = fcntl(client, F_SETFL, O_NONBLOCK);
         if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't set non blocking mode for client.\n");
+                prelude_log(PRELUDE_LOG_ERR, "could not set non blocking mode for client: %s.\n", strerror(errno));
                 return -1;
         }
-
-        cdata->fd = prelude_io_new();
-        if ( ! cdata->fd ) 
+        
+        fcntl(client, F_SETFD, fcntl(client, F_GETFD) | FD_CLOEXEC);
+        
+        ret = prelude_io_new(&cdata->fd);
+        if ( ret < 0 ) 
                 return -1;
 
-        prelude_io_set_socket_io(cdata->fd, client);
+        prelude_io_set_sys_io(cdata->fd, client);
                
         cdata->msg = NULL;
-        cdata->is_ssl = 0;
-        cdata->is_authenticated = 0;
+        cdata->state = 0;
         
         return 0;
 }
@@ -462,7 +356,7 @@ static int accept_connection(server_generic_t *server, server_generic_client_t *
 {
         int sock;
         socklen_t addrlen;
-
+        
 #ifndef HAVE_IPV6
         struct sockaddr_in addr;
 #else
@@ -473,23 +367,34 @@ static int accept_connection(server_generic_t *server, server_generic_client_t *
         
         sock = accept(server->sock, (struct sockaddr *) &addr, &addrlen);
         if ( sock < 0 ) {
-                log(LOG_ERR, "accept returned an error.\n");
+                prelude_log(PRELUDE_LOG_ERR, "accept error: %s.\n", strerror(errno));
                 return -1;
         }
-        
+                
         if ( server->sa->sa_family == AF_UNIX )
-                cdata->addr = strdup("unix");
-        else {         
+                cdata->addr = strdup(((struct sockaddr_un *) server->sa)->sun_path);
+        else {
                 void *in_addr;
                 char out[128];
                 const char *str;
                 struct sockaddr *sa = (struct sockaddr *) &addr;
-                
-                in_addr = prelude_inet_sockaddr_get_inaddr(sa);
+
+#ifdef HAVE_IPV6
+                cdata->port = ntohs(addr.sin6_port);
+#else
+		cdata->port = ntohs(addr.sin_port);
+#endif
+                in_addr = prelude_sockaddr_get_inaddr(sa);
+                if ( ! in_addr ) {
+                        close(sock);
+                        return -1;
+                }
                 
                 str = inet_ntop(sa->sa_family, in_addr, out, sizeof(out));
-                if ( str )                
-                        cdata->addr = strdup(str);
+                if ( str ) {
+                        snprintf(out + strlen(out), sizeof(out) - strlen(out), ":%d", cdata->port);
+                        cdata->addr = strdup(out);
+                }
         }
 
         if ( ! cdata->addr ) {
@@ -505,20 +410,20 @@ static int accept_connection(server_generic_t *server, server_generic_client_t *
 
 
 
-static int handle_connection(server_generic_t *server, prelude_msg_t *cfgmsg) 
+static int handle_connection(server_generic_t *server) 
 {
         int ret, client;
         server_generic_client_t *cdata;
         
-        cdata = malloc(server->clientlen);
+        cdata = calloc(1, server->clientlen);
         if ( ! cdata ) {
-                log(LOG_ERR, "memory exhausted.\n");
+                prelude_log(PRELUDE_LOG_ERR, "memory exhausted.\n");
                 return -1;
         }
 
         client = accept_connection(server, cdata);                
         if ( client < 0 ) {
-                log(LOG_ERR, "couldn't accept connection.\n");
+                prelude_log(PRELUDE_LOG_ERR, "couldn't accept connection.\n");
                 free(cdata);
                 return -1;
         }
@@ -529,20 +434,10 @@ static int handle_connection(server_generic_t *server, prelude_msg_t *cfgmsg)
                 close(client);
                 return -1;
         }
-                
-        ret = prelude_msg_write(cfgmsg, cdata->fd);
-        if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't send configuration message.\n");
-                prelude_io_close(cdata->fd);
-                prelude_io_destroy(cdata->fd);
-                free(cdata->addr);
-                free(cdata);
-                return -1;
-        }
-
+        
         ret = server_logic_process_requests(server->logic, (server_logic_client_t *) cdata);
         if ( ret < 0 ) {
-                log(LOG_ERR, "queueing client FD for server logic processing failed.\n");
+                prelude_log(PRELUDE_LOG_ERR, "queueing client FD for server logic processing failed.\n");
                 prelude_io_close(cdata->fd);
                 prelude_io_destroy(cdata->fd);
                 free(cdata->addr);
@@ -563,20 +458,16 @@ static int handle_connection(server_generic_t *server, prelude_msg_t *cfgmsg)
  */
 static int wait_connection(server_generic_t **server, size_t nserver)
 {
-        int i, active_fd;
-        prelude_msg_t *cfgmsg;
+        size_t i;
+        int active_fd;
         struct pollfd pfd[nserver];
-                
-        cfgmsg = generate_config_message();
-        if ( ! cfgmsg )
-                return -1;
-
+        
         for ( i = 0; i < nserver; i++ ) {                
                 pfd[i].events = POLLIN;
                 pfd[i].fd = server[i]->sock;
         } 
         
-        while ( 1 ) {
+        while ( continue_processing ) {
 
                 active_fd = poll(pfd, nserver, -1);                
                 if ( active_fd < 0 )
@@ -585,10 +476,12 @@ static int wait_connection(server_generic_t **server, size_t nserver)
                 for ( i = 0; i < nserver && active_fd > 0; i++ ) {
                         if ( pfd[i].revents & POLLIN ) {
                                 active_fd--;
-                                handle_connection(server[i], cfgmsg);
+                                handle_connection(server[i]);
                         }
                 }
         }
+
+        return 0;
 }
 
 
@@ -602,13 +495,13 @@ static int generic_server(int sock, struct sockaddr *addr, size_t alen)
         
         ret = bind(sock, addr, alen);
         if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't bind to socket.\n");
+                prelude_log(PRELUDE_LOG_ERR, "could not bind socket: %s.\n", strerror(errno));
                 return -1;
         }
         
         ret = listen(sock, 10);
         if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't listen on socket.\n");
+                prelude_log(PRELUDE_LOG_ERR, "could not listen on socket: %s.\n", strerror(errno));
                 return -1;
         }
         
@@ -631,15 +524,27 @@ static int generic_server(int sock, struct sockaddr *addr, size_t alen)
 static int is_unix_socket_already_used(int sock, struct sockaddr_un *sa, int addrlen) 
 {
         int ret;
-        
-        ret = access(sa->sun_path, F_OK);
+        struct stat st;
+
+        /*
+         * Minor check for operating system where bind() will incorrectly follow symlink.
+         * Note that there is a race condition between the stat() and the bind call.
+         */
+        ret = stat(sa->sun_path, &st);
         if ( ret < 0 )
-                return 0;
+                return FALSE;
+
+        if ( ! S_ISSOCK(st.st_mode) ) {
+                prelude_log(PRELUDE_LOG_WARN, "%s already exist and is not an UNIX socket: please check.\n",
+                            sa->sun_path);
+                return -1;
+        }
         
         ret = connect(sock, (struct sockaddr *) sa, addrlen);
         if ( ret == 0 ) {
-                log(LOG_INFO, "Prelude Manager UNIX socket is already used. Exiting.\n");
-                return 1;
+                prelude_log(PRELUDE_LOG_WARN, "%s UNIX socket is already in use. Exiting.\n",
+                            sa->sun_path);
+                return TRUE;
         }
         
         /*
@@ -648,7 +553,7 @@ static int is_unix_socket_already_used(int sock, struct sockaddr_un *sa, int add
          */
         ret = unlink(sa->sun_path);
         if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't delete UNIX socket.\n");
+                prelude_log(PRELUDE_LOG_ERR, "couldn't delete UNIX socket.\n");
                 return -1;
         }
         
@@ -667,17 +572,17 @@ static int unix_server_start(server_generic_t *server)
         
         server->sock = socket(AF_UNIX, SOCK_STREAM, 0);
         if ( server->sock < 0 ) {
-                log(LOG_ERR, "couldn't create socket.\n");
+                prelude_log(PRELUDE_LOG_ERR, "error creating UNIX socket: %s.\n", strerror(errno));
 		return -1;
 	}
-        
-        ret = is_unix_socket_already_used(server->sock, sa, server->sa_len);
+
+        ret = is_unix_socket_already_used(server->sock, sa, server->slen);
         if ( ret == 1 || ret < 0  ) {
                 close(server->sock);
                 return -1;
         }
 
-        ret = generic_server(server->sock, server->sa, server->sa_len);
+        ret = generic_server(server->sock, server->sa, server->slen);
         if ( ret < 0 ) {
                 close(server->sock);
                 return -1;
@@ -689,7 +594,7 @@ static int unix_server_start(server_generic_t *server)
          */
         ret = chmod(sa->sun_path, S_IRWXU|S_IRWXG|S_IRWXO);
         if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't set permission for UNIX socket.\n");
+                prelude_log(PRELUDE_LOG_ERR, "could not set permission on UNIX socket: %s.\n", strerror(errno));
                 return -1;
         }
         
@@ -702,26 +607,25 @@ static int unix_server_start(server_generic_t *server)
 /*
  *
  */
-static int inet_server_start(server_generic_t *server,
-                             const char *saddr, struct sockaddr *addr, socklen_t addrlen) 
+static int inet_server_start(server_generic_t *server, struct sockaddr *addr, socklen_t addrlen) 
 {
         int ret, on = 1;
         
         server->sock = socket(server->sa->sa_family, SOCK_STREAM, IPPROTO_TCP);
         if ( server->sock < 0 ) {
-                log(LOG_ERR, "couldn't create socket.\n");
+                prelude_log(PRELUDE_LOG_ERR, "could not create socket: %s.\n", strerror(errno));
                 return -1;
         }
         
         ret = setsockopt(server->sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(int));
         if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't set SO_REUSEADDR socket option.\n");
+                prelude_log(PRELUDE_LOG_ERR, "could not set SO_REUSEADDR socket option: %s.\n", strerror(errno));
                 goto err;
         }
 
         ret = setsockopt(server->sock, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(int));
         if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't set SO_KEEPALIVE socket option.\n");
+                prelude_log(PRELUDE_LOG_ERR, "could not set SO_KEEPALIVE socket option: %s.\n", strerror(errno));
                 goto err;
         }
         
@@ -729,12 +633,6 @@ static int inet_server_start(server_generic_t *server,
         if ( ret < 0 )
                 goto err;
 
-#ifdef HAVE_SSL
-        ret = ssl_init_server();
-	if ( ret < 0 )
-                goto err;
-#endif
-
         return 0;
 
  err:
@@ -744,51 +642,88 @@ static int inet_server_start(server_generic_t *server,
 
 
 
-
-static int resolve_addr(server_generic_t *server, const char *addr, uint16_t port) 
+static prelude_bool_t is_unix_addr(const char **out, const char *addr)
 {
         int ret;
-        prelude_addrinfo_t *ai, hints;
+        const char *ptr;
+
+        if ( ! addr )
+                return FALSE;
+        
+        ret = strncmp(addr, "unix", 4);
+        if ( ret != 0 )
+                return FALSE;
+        
+        ptr = strchr(addr, ':');        
+        *out = (ptr && *(ptr + 1)) ? ptr + 1 : prelude_connection_get_default_socket_filename();
+        
+        return TRUE;
+}
+
+
+
+static int do_getaddrinfo(struct addrinfo **ai, const char *addr, unsigned int port)
+{
+        int ret;
+        struct addrinfo hints;
         char service[sizeof("00000")];
 
         memset(&hints, 0, sizeof(hints));
+        snprintf(service, sizeof(service), "%u", port);
         
         hints.ai_family = PF_UNSPEC;
         hints.ai_socktype = SOCK_STREAM;
         hints.ai_protocol = IPPROTO_TCP;
-
-        snprintf(service, sizeof(service), "%u", port);
-        
-        ret = prelude_inet_getaddrinfo(addr, service, &hints, &ai);        
-        if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't resolve %s.\n", addr);
+                
+        ret = getaddrinfo(addr, service, &hints, ai);
+        if ( ret != 0 ) {
+                prelude_log(PRELUDE_LOG_WARN, "could not resolve %s: %s.\n",
+                            addr, (ret == EAI_SYSTEM) ? strerror(errno) : gai_strerror(ret));
                 return -1;
         }
 
-        ret = prelude_inet_addr_is_loopback(ai->ai_family, prelude_inet_sockaddr_get_inaddr(ai->ai_addr));
-        if ( ret == 0 ) {
-                ai->ai_family = AF_UNIX;
-                ai->ai_addrlen = sizeof(struct sockaddr_un);
-        }
+        return 0;
+}
+
+
+
+static int resolve_addr(server_generic_t *server, const char *addr, unsigned int port) 
+{
+        struct addrinfo *ai;
+        const char *unixpath = NULL;
+        int ret, ai_family, ai_addrlen;
         
-        server->sa = malloc(ai->ai_addrlen);
-        if ( ! server->sa ) {
-                log(LOG_ERR, "memory exhausted.\n");
-                prelude_inet_freeaddrinfo(ai);
-                return -1;
+        if ( is_unix_addr(&unixpath, addr) ) {
+                ai_family = AF_UNIX;
+                ai_addrlen = sizeof(struct sockaddr_un);
         }
 
-        server->sa_len = ai->ai_addrlen;
-        server->sa->sa_family = ai->ai_family;
-        
-        if ( ai->ai_family != AF_UNIX )
-                memcpy(server->sa, ai->ai_addr, ai->ai_addrlen);
         else {
+                ret = do_getaddrinfo(&ai, addr, port);
+                if ( ret < 0 )
+                        return -1;
+
+                ai_family = ai->ai_family;
+                ai_addrlen = ai->ai_addrlen;
+        }
+
+        server->sa = malloc(ai_addrlen);
+        if ( ! server->sa ) {
+                prelude_log(PRELUDE_LOG_ERR, "memory exhausted.\n");
+                freeaddrinfo(ai);
+                return -1;
+        }
+
+        server->slen = ai_addrlen;
+        server->sa->sa_family = ai_family;
+                
+        if ( ai_family != AF_UNIX ) {
+                memcpy(server->sa, ai->ai_addr, ai->ai_addrlen);
+                freeaddrinfo(ai);
+        } else {
                 struct sockaddr_un *un = (struct sockaddr_un *) server->sa;
-                prelude_get_socket_filename(un->sun_path, sizeof(un->sun_path), port);
+                strncpy(un->sun_path, unixpath, sizeof(un->sun_path));
         }
-        
-        prelude_inet_freeaddrinfo(ai);
 
         return 0;
 }
@@ -799,64 +734,76 @@ static int resolve_addr(server_generic_t *server, const char *addr, uint16_t por
 /*
  *
  */
-server_generic_t *server_generic_new(const char *saddr, uint16_t port,
-                                     size_t clientlen, server_generic_accept_func_t *acceptf,
-                                     server_generic_read_func_t *readf, server_generic_close_func_t *closef)
+server_generic_t *server_generic_new(size_t clientlen, server_generic_accept_func_t *acceptf,
+                                     server_generic_read_func_t *readf,
+                                     server_generic_write_func_t *writef,
+                                     server_generic_close_func_t *closef)
 {
-        int ret;
-        char out[128];
-        void *in_addr;
         server_generic_t *server;
-        
+                
         server = malloc(sizeof(*server));
         if ( ! server ) {
-                log(LOG_ERR, "memory exhausted.\n");
+                prelude_log(PRELUDE_LOG_ERR, "memory exhausted.\n");
                 return NULL;
         }
 
         server->read = readf;
+        server->write = writef;
         server->accept = acceptf;
         server->close = closef;
         server->clientlen = clientlen;
-
-        ret = resolve_addr(server, saddr, port);
-        if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't resolve %s.\n", saddr);
-                return NULL;
-        }
         
-        server->logic = server_logic_new(server, read_connection_cb, NULL, close_connection_cb);
+        server->logic = server_logic_new(server, read_connection_cb, write_connection_cb, close_connection_cb);
         if ( ! server->logic ) {
-                log(LOG_ERR, "couldn't initialize server pool.\n");
+                prelude_log(PRELUDE_LOG_WARN, "couldn't initialize server pool.\n");
                 free(server);
                 return NULL;
         }
         
-        if ( server->sa->sa_family == AF_UNIX )
-                ret = unix_server_start(server);
-        else 
-                ret = inet_server_start(server, saddr, server->sa, server->sa_len);
-        
-        if ( ret < 0 ) {
-                server_logic_stop(server->logic);
-                free(server);
-                return NULL;
-        }
-
-        if ( server->sa->sa_family == AF_UNIX )
-                snprintf(out, sizeof(out), "unix socket");
-        else {
-                assert(in_addr = prelude_inet_sockaddr_get_inaddr(server->sa));
-                prelude_inet_ntop(server->sa->sa_family, in_addr, out, sizeof(out));
-        }
-        
-        log(LOG_INFO, "- sensors server started (listening on %s port %d).\n", out, port);
-
         return server;
 }
 
 
 
+int server_generic_bind(server_generic_t *server, const char *saddr, unsigned int port)
+{
+        int ret;
+        char out[128];
+        void *in_addr;
+        
+        ret = resolve_addr(server, saddr, port);
+        if ( ret < 0 )
+                return ret;
+        
+        if ( server->sa->sa_family == AF_UNIX )
+                ret = unix_server_start(server);
+        else 
+                ret = inet_server_start(server, server->sa, server->slen);
+        
+        if ( ret < 0 ) {
+                server_logic_stop(server->logic);
+                free(server->sa);
+                free(server);
+                return -1;
+        }
+                
+        fcntl(server->sock, F_SETFD, fcntl(server->sock, F_GETFD) | FD_CLOEXEC);
+        
+        if ( server->sa->sa_family == AF_UNIX )
+                prelude_log(PRELUDE_LOG_INFO, "- server started (listening on %s).\n",
+                            ((struct sockaddr_un *) server->sa)->sun_path);
+        else {
+                in_addr = prelude_sockaddr_get_inaddr(server->sa);
+                assert(in_addr);
+                
+                inet_ntop(server->sa->sa_family, in_addr, out, sizeof(out));
+                prelude_log(PRELUDE_LOG_INFO, "- server started (listening on %s port %u).\n", out, port);
+        }
+                
+
+        return 0;
+}
+
 
 void server_generic_start(server_generic_t **server, size_t nserver) 
 {
@@ -865,6 +812,14 @@ void server_generic_start(server_generic_t **server, size_t nserver)
 
 
 
+
+void server_generic_stop(server_generic_t *server)
+{
+        continue_processing = 0;
+}
+
+
+
 void server_generic_close(server_generic_t *server) 
 {
         close(server->sock);
@@ -877,10 +832,79 @@ void server_generic_close(server_generic_t *server)
 
 
 
+void server_generic_process_requests(server_generic_t *server, server_generic_client_t *client)
+{
+        server_logic_process_requests(server->logic, (server_logic_client_t *) client);
+}
 
 
 
+void server_generic_log_client(server_generic_client_t *cnx, prelude_log_t priority, const char *fmt, ...)
+{
+        va_list ap;
+        int ret = 0;
+        char buf[1024];
+        
+        if ( cnx->ident && cnx->permission_string ) {
+                ret = snprintf(buf, sizeof(buf), " 0x%" PRELUDE_PRIx64 " %s]: ", cnx->ident, cnx->permission_string);
+                if ( ret < 0 || ret >= sizeof(buf) )
+                        return;
+        } else {
+                ret = snprintf(buf, sizeof(buf), "]: ");
+                if ( ret < 0 || ret >= sizeof(buf) )
+                        return;
+        }
+                
+        va_start(ap, fmt);
+        vsnprintf(buf + ret, sizeof(buf) - ret, fmt, ap);
+        va_end(ap);
+        
+        prelude_log(priority, "[%s%s", cnx->addr, buf);
+}
 
 
 
+void server_generic_client_set_analyzerid(server_generic_client_t *client, uint64_t analyzerid)
+{
+        client->ident = analyzerid;
+}
 
+
+void server_generic_client_set_state(server_generic_client_t *client, int state)
+{
+        client->state = state;
+}
+
+
+int server_generic_client_get_state(server_generic_client_t *client)
+{
+        return client->state;
+}
+
+
+
+int server_generic_client_set_permission(server_generic_client_t *client, prelude_connection_permission_t permission)
+{
+        int ret;
+        prelude_string_t *out;
+
+        ret = prelude_string_new(&out);
+        if ( ret < 0 )
+                return ret;
+
+        ret = prelude_connection_permission_to_string(permission, out);
+        if ( ret < 0 ) {
+                prelude_string_destroy(out);
+                return ret;
+        }
+
+        ret = prelude_string_get_string_released(out, &client->permission_string);
+        prelude_string_destroy(out);
+        
+        if ( ret < 0 )
+                return ret;
+
+        client->permission = permission;
+
+        return 0;
+}
