@@ -1,9 +1,9 @@
 /*****
 *
-* Copyright (C) 2001, 2002, 2003 Yoann Vandoorselaere <yoann@prelude-ids.org>
-* All Rights Reserved
+* Copyright (C) 2001, 2002, 2003, 2004, 2005 PreludeIDS Technologies. All Rights Reserved.
+* Author: Yoann Vandoorselaere <yoann.v@prelude-ids.com>
 *
-* This file is part of the Prelude program.
+* This file is part of the Prelude-Manager program.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
@@ -21,20 +21,21 @@
 *
 *****/
 
+#include "libmissing.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/poll.h>
 #include <pthread.h>
 #include <assert.h>
-#include <inttypes.h>
 #include <errno.h>
 #include <signal.h>
 #include <sys/time.h>
 
-#include <libprelude/list.h>
+#include <libprelude/prelude-inttypes.h>
+#include <libprelude/prelude-list.h>
 #include <libprelude/prelude-log.h>
 #include <libprelude/prelude-io.h>
-#include <libprelude/threads.h>
 
 #include "server-logic.h"
 
@@ -60,12 +61,7 @@
  */
 #define DEFAULT_MAX_FD_BY_THREAD 100
 
-
-#ifdef DEBUG
- #define dprint(args...) fprintf(stderr, args)
-#else
- #define dprint(args...) do { } while(0)
-#endif
+#define dprint(args...) prelude_log_debug(10, args)
 
 
 struct server_logic_client {
@@ -75,7 +71,7 @@ struct server_logic_client {
 
 
 typedef struct server_fd_set {
-        struct list_head list;
+        prelude_list_t list;
 
         /*
          * Thread handling this set of file descriptor.
@@ -93,7 +89,7 @@ typedef struct server_fd_set {
          * Index used to address free client data / pfd.
          */
         volatile sig_atomic_t used_index;
-        
+
         server_logic_t *parent;
 
         /*
@@ -133,8 +129,8 @@ struct server_logic {
          * List of connection set associated with this server.
          */
         pthread_mutex_t mutex;
-        struct list_head free_set_list;
-        struct list_head used_set_list;
+        prelude_list_t free_set_list;
+        prelude_list_t  used_set_list;
 
         volatile sig_atomic_t continue_processing;
 };
@@ -171,7 +167,7 @@ static int setup_sigusr1_action(sigset_t *set)
 
         ret = sigaction(SIGUSR1, &act, NULL);
         if ( ret < 0 ) 
-                log(LOG_ERR, "failed to register thread handler for SIGUSR1.\n");
+                prelude_log(PRELUDE_LOG_ERR, "failed to register thread handler for SIGUSR1.\n");
 
         return ret;
 }
@@ -261,8 +257,8 @@ static inline void add_set_to_free_list(server_logic_t *server, server_fd_set_t
         dprint("thread=%ld, Adding to list, used_index=%d, thread_num=%d.\n",
                set->thread, set->used_index, server->thread_num);
 
-        list_del(&set->list);
-        list_add_tail(&set->list, &server->free_set_list);
+        prelude_list_del(&set->list);
+        prelude_list_add_tail(&server->free_set_list, &set->list);
 }
 
 
@@ -279,20 +275,21 @@ static void update_fd_set_status(server_logic_t *server, server_fd_set_t *set)
 
 
 
-static void remove_connection(server_fd_set_t *set, int cnx_key) 
+static int remove_connection(server_fd_set_t *set, int cnx_key) 
 {
+        int ret;
         server_logic_t *server = set->parent;
 
         dprint("removing connection\n");
-        
+                
         /*
          * Close the file descriptor associated with this set.
          */
-        pthread_mutex_lock(&set->client[cnx_key]->mutex);
-        set->client[cnx_key]->key = -1;
-        pthread_mutex_unlock(&set->client[cnx_key]->mutex);
-
-        server->close(server->sdata, set->client[cnx_key]);
+        ret = server->close(server->sdata, set->client[cnx_key]);
+        if ( ret < 0 ) {
+                dprint("remove connection not completed\n");
+                return ret;
+        }
         
         /*
          * As of now we need to obtain the server lock because
@@ -304,6 +301,8 @@ static void remove_connection(server_fd_set_t *set, int cnx_key)
         update_fd_set_status(server, set);
         
         pthread_mutex_unlock(&server->mutex);
+
+        return 0;
 }
 
 
@@ -322,42 +321,51 @@ static void add_connection(server_logic_t *server, server_fd_set_t *set, server_
         /*
          * If the set is full, remove it from out list.
          * The list should be locked when this function is called !
-         */        
+         */
         if ( set->used_index == server->thread_max_fd ) {
                 dprint("[%ld][%p] Max connection for this thread reached (%d).\n", set->thread, set, set->used_index);
-                list_del(&set->list);
-                list_add_tail(&set->list, &server->used_set_list);
+                prelude_list_del(&set->list);
+                prelude_list_add_tail(&server->used_set_list, &set->list);
         }
 }
 
 
-
-
 static int handle_fd_event(server_fd_set_t *set, int cnx_key) 
 {
-        int ret = -2;
+        int ret = 0;
+        prelude_bool_t got_event = FALSE;
         
         assert(set->client[cnx_key]->key == cnx_key);
-                
-        if ( ret != -1 && set->pfd[cnx_key].revents & (POLLERR|POLLHUP|POLLNVAL) ) {
-                dprint("thread=%ld: key=%d, fd=%d: Hanging up.\n", pthread_self(), cnx_key, set->pfd[cnx_key].fd);
-                ret = -1;
-        }
         
         if ( set->pfd[cnx_key].events & POLLIN && set->pfd[cnx_key].revents & POLLIN ) {                
+                got_event = TRUE;
                 ret = set->parent->read(set->parent->sdata, set->client[cnx_key]);
                 dprint("thread=%ld: key=%d, fd=%d: Data available (ret=%d)\n", pthread_self(), cnx_key, set->pfd[cnx_key].fd, ret);
         }
 
-        if ( ret != -1 && set->pfd[cnx_key].events & POLLOUT && set->pfd[cnx_key].revents & POLLOUT ) {
+        /*
+         * POLLHUP and POLLOUT are mutually exclusive.
+         */
+        if ( ! got_event && ret >= 0 && set->pfd[cnx_key].revents & (POLLERR|POLLHUP|POLLNVAL) ) {
+                got_event = TRUE;
+                dprint("thread=%ld: key=%d, fd=%d: Hanging up.\n", pthread_self(), cnx_key, set->pfd[cnx_key].fd);
+                ret = -1; /* trigger remove_connection() */
+        }
+
+        else if ( ret >= 0 && set->pfd[cnx_key].events & POLLOUT && set->pfd[cnx_key].revents & POLLOUT ) {
+                got_event = TRUE;
                 dprint("thread=%ld: key=%d, fd=%d: Output possible.\n", pthread_self(), cnx_key, set->pfd[cnx_key].fd);
                 ret = set->parent->write(set->parent->sdata, set->client[cnx_key]);
         }
+
+        if ( ! got_event )
+                return -1;
         
         if ( ret < 0 ) {
-                if ( ret == -2 ) 
-                        return -1;
-
+                if ( ret == -2 )
+                        /* callback asked to stop processing connection */
+                        return 0;
+                
                 remove_connection(set, cnx_key);
         }
                 
@@ -377,7 +385,7 @@ static void poll_fd_set(server_fd_set_t *set)
                 if ( errno == EINTR ) 
                         return;
                 
-                log(LOG_ERR, "error polling FDs set.\n");
+                prelude_log(PRELUDE_LOG_ERR, "error polling FDs set.\n");
         }
                 
         for ( i = 0; i < set->parent->thread_max_fd && active_fd > 0; i++ ) {
@@ -399,15 +407,17 @@ static void poll_fd_set(server_fd_set_t *set)
 static void destroy_fd_set(server_logic_t *server, server_fd_set_t *set) 
 {
         int i;
-
-        list_del(&set->list);
+        
+        dprint("killing thread %ld on exit request.\n", set->thread);
+        
+        prelude_list_del(&set->list);
         server->thread_num--;
 
         pthread_mutex_unlock(&set->parent->mutex);
         
         for ( i = 0; i < set->used_index; i++ )
                 server->close(server->sdata, set->client[i]);
-
+        
         pthread_cond_destroy(&set->startup_cond);
         pthread_mutex_destroy(&set->startup_mutex);
         
@@ -463,8 +473,7 @@ static void *child_reader(void *ptr)
                 
                 poll_fd_set(set);
         }
-
-        dprint("killing thread %ld on exit request.\n", set->thread);
+        
         pthread_exit(NULL);
 
         return NULL; /* not needed, but avoid a warning. */
@@ -479,20 +488,20 @@ static server_fd_set_t *create_fd_set(server_logic_t *server)
         
         new = malloc(sizeof(*new));
         if ( ! new ) {
-                log(LOG_ERR, "memory exhausted.\n");
+                prelude_log(PRELUDE_LOG_ERR, "memory exhausted.\n");
                 return NULL;
         }
 
         new->pfd = malloc(sizeof(struct pollfd) * server->thread_max_fd);
         if ( ! new->pfd ) {
-                log(LOG_ERR, "memory exhausted.\n");
+                prelude_log(PRELUDE_LOG_ERR, "memory exhausted.\n");
                 free(new);
                 return NULL;
         }
-        
+
         new->client = malloc(sizeof(new->client) * server->thread_max_fd);
         if ( ! new->client ) {
-                log(LOG_ERR, "memory exhausted.\n");
+                prelude_log(PRELUDE_LOG_ERR, "memory exhausted.\n");
                 free(new->pfd);
                 free(new);
                 return NULL;
@@ -504,14 +513,14 @@ static server_fd_set_t *create_fd_set(server_logic_t *server)
         
         pthread_cond_init(&new->startup_cond, NULL);
         pthread_mutex_init(&new->startup_mutex, NULL);
-
+        
         for ( i = 0; i < server->thread_max_fd; i++ ) {
                 new->pfd[i].fd = -1;
                 new->client[i] = NULL;
         }
 
         pthread_mutex_lock(&server->mutex);
-        list_add_tail(&new->list, &server->free_set_list);
+        prelude_list_add_tail(&server->free_set_list, &new->list);
         pthread_mutex_unlock(&server->mutex);
         
         return new;
@@ -527,8 +536,8 @@ static int start_fd_set_thread(server_logic_t *server, server_fd_set_t *set)
         
         ret = pthread_create(&set->thread, NULL, &child_reader, set);                
         if ( ret < 0 ) {
-                log(LOG_ERR, "couldn't create thread.\n");
                 pthread_mutex_unlock(&set->startup_mutex);
+                prelude_log(PRELUDE_LOG_ERR, "couldn't create thread.\n");
                 return -1;
         }
 
@@ -564,8 +573,6 @@ int server_logic_process_requests(server_logic_t *server, server_logic_client_t
 {
         int ret;
         server_fd_set_t *set;
-
-        pthread_mutex_init(&client->mutex, NULL);
         
         /*
          * Hold the list lock until we add the connection.
@@ -580,8 +587,8 @@ int server_logic_process_requests(server_logic_t *server, server_logic_client_t
                 return -1;
         }
         
-        if ( ! list_empty(&server->free_set_list) ) {
-                set = list_entry(server->free_set_list.next, server_fd_set_t, list);
+        if ( ! prelude_list_is_empty(&server->free_set_list) ) {
+                set = prelude_list_entry(server->free_set_list.next, server_fd_set_t, list);
                 add_connection(server, set, client);
 
                 pthread_mutex_unlock(&server->mutex);
@@ -678,9 +685,9 @@ void server_logic_set_max_fd_by_thread(server_logic_t *server, unsigned int max)
  * @client:
  *
  */
-void server_logic_remove_client(server_logic_client_t *client) 
+int server_logic_remove_client(server_logic_client_t *client) 
 {
-        remove_connection(client->set, client->key);
+        return remove_connection(client->set, client->key);
 }
 
 
@@ -703,8 +710,8 @@ server_logic_t *server_logic_new(void *sdata, server_logic_read_t *s_read,
         if ( ! new )
                 return NULL;
 
-        INIT_LIST_HEAD(&new->free_set_list);
-        INIT_LIST_HEAD(&new->used_set_list);
+        prelude_list_init(&new->free_set_list);
+        prelude_list_init(&new->used_set_list);
         pthread_mutex_init(&new->mutex, NULL);
 
         new->sdata = sdata;
@@ -730,18 +737,8 @@ void server_logic_notify_write_enable(server_logic_client_t *fd)
 {
         server_fd_set_t *set = fd->set;
         
-        pthread_mutex_lock(&fd->mutex);
-        
-        if ( fd->key < 0 ) {
-                pthread_mutex_unlock(&fd->mutex);
-                return;
-
-        }
-        
         set->pfd[fd->key].events |= POLLOUT;
         pthread_kill(set->thread, SIGUSR1);
-
-        pthread_mutex_unlock(&fd->mutex);
 }
  
 
